#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

time_map = {
    'picosecond'  : 'ps',
    'picoseconds' : 'ps',
    'picosec'     : 'ps',
    'picosecs'    : 'ps',
    'ps'          : 'ps',

    'nanosecond'  : 'ns',
    'nanoseconds' : 'ns',
    'nanosecs'    : 'ns',
    'nanosec'     : 'ns',
    'ns'          : 'ns',

    'microsecond'  : 'µs',
    'microseconds' : 'µs',
    'microsec'     : 'µs',
    'microsecs'    : 'µs',
    'µs'           : 'µs',

    'millisecond'   : 'ms',
    'milliseconds'  : 'ms',
    'millisec'      : 'ms',
    'millisecs'     : 'ms',
    'ms'            : 'ms',

    'second'  : 's',
    'seconds' : 's',
    'sec'     : 's',
    'secs'    : 's',
    's'       : 's',
    
    'minute'  : 'min',
    'minutes' : 'min',
    'min'     : 'min',
    'mins'    : 'min',
    'm'       : 'min',

    'h'     : 'h',
    'hour'  : 'h',
    'hours' : 'h',

    'd'    : 'day',
    'day'  : 'day',
    'days' : 'day',

    'mon'    : 'mon',
    'month'  : 'mon',
    'months' : 'mon',
    'mons'   : 'mon',

    'y'    : 'yr',
    'yr'   : 'yr',
    'year' : 'yr',
    'years': 'yr',
    'yrs'  : 'yr',
}


